//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/PointwiseAxisItem.cpp
//! @brief     Implements pointwise axis item
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Device/Coord/ICoordSystem.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "Device/IO/ReadWriteINT.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {

namespace Tag {

const QString NativeUnits("NativeUnits");
const QString BinaryData("BinaryData");

} // namespace Tag

namespace Attrib {

const QString value("value");

}

} // namespace

PointwiseAxisItem::PointwiseAxisItem(QObject* parent)
    : BasicAxisItem(parent)
    , m_instrumentItem(nullptr)
    , m_nativeAxisUnits("nbins")
{
}

PointwiseAxisItem::~PointwiseAxisItem() = default;

void PointwiseAxisItem::init(const IAxis& axis, const QString& units_label)
{
    m_axis = std::unique_ptr<IAxis>(axis.clone());
    m_nativeAxisUnits = units_label;
}

void PointwiseAxisItem::setOwningInstrument(const InstrumentItem* instrument)
{
    // only specular instruments are relevant
    m_instrumentItem = dynamic_cast<const SpecularInstrumentItem*>(instrument);
}

const IAxis* PointwiseAxisItem::axis() const
{
    return m_axis.get();
}

QString PointwiseAxisItem::nativeAxisUnits() const
{
    return m_nativeAxisUnits;
}

std::unique_ptr<IAxis> PointwiseAxisItem::createAxis(double scale) const
{
    if (!checkValidity())
        return nullptr;

    auto* const converter = m_instrumentItem->createCoordSystem();
    std::unique_ptr<IAxis> converted_axis(converter->createConvertedAxis(0, Coords::DEGREES));

    // applying scaling
    std::vector<double> centers = converted_axis->binCenters();
    for (double& e : centers)
        e *= scale;

    return std::make_unique<PointwiseAxis>(converted_axis->axisName(), std::move(centers));
}

QByteArray PointwiseAxisItem::serializeBinaryData() const
{
    if (!m_axis)
        return {};
    Datafield axisData({m_axis->clone()});

    std::stringstream ss;
    ReadWriteINT().writeDatafield(axisData, ss);
    return QByteArray(ss.str().c_str(), static_cast<int>(ss.str().size()));
}

void PointwiseAxisItem::deserializeBinaryData(const QByteArray& data)
{
    if (data.isEmpty())
        return;

    std::istringstream str(data.toStdString());
    std::unique_ptr<Datafield> d(ReadWriteINT().readDatafield(str));
    m_axis = std::unique_ptr<IAxis>(d->axis(0).clone());
}

void PointwiseAxisItem::writeTo(QXmlStreamWriter* writer) const
{
    BasicAxisItem::writeTo(writer);

    // axis binary data
    QByteArray a = serializeBinaryData();
    if (!a.isEmpty()) {
        writer->writeStartElement(Tag::BinaryData);
        writer->writeCharacters(a.toBase64());
        writer->writeEndElement();
    }

    // native units
    writer->writeStartElement(Tag::NativeUnits);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_nativeAxisUnits);
    writer->writeEndElement();
}

void PointwiseAxisItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // axis binary data
        if (reader->name() == Tag::BinaryData) {
            QString valueAsBase64 = reader->readElementText(QXmlStreamReader::SkipChildElements);
            const auto data = QByteArray::fromBase64(valueAsBase64.toLatin1());
            deserializeBinaryData(data);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::BinaryData);

            // native units
        } else if (reader->name() == Tag::NativeUnits) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_nativeAxisUnits);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::NativeUnits);

            // members from base class
        } else {
            BasicAxisItem::readBasicTags(reader);
        }
    }
}

bool PointwiseAxisItem::checkValidity() const
{
    return m_axis && m_instrumentItem && nativeAxisUnits() != "nbins";
}

void PointwiseAxisItem::updateIndicators()
{
    if (!checkValidity())
        return;

    auto* const converter = m_instrumentItem->createCoordSystem();
    setMin(converter->calculateMin(0, Coords::DEGREES));
    setMax(converter->calculateMax(0, Coords::DEGREES));
    setBinCount(static_cast<int>(m_axis->size()));
}
