//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleLayoutItem.cpp
//! @brief     Implements class ParticleLayoutItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/CatSample/ItemWithParticlesCatalog.h"
#include "GUI/Model/Sample/CompoundItem.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/MesocrystalItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Support/XML/Serialize.h"

namespace {
namespace Tag {

const QString OwnDensity("OwnDensity");
const QString Particles("Particles");
const QString InterferenceFunction("InterferenceFunction");

} // namespace Tag
} // namespace

ParticleLayoutItem::ParticleLayoutItem(const MaterialItems* materials)
    : m_materials(materials)
{
    m_ownDensity.init("Total particle density",
                      "Number of particles per area (particle surface density).\n "
                      "Should be defined for disordered and 1d-ordered particle collections.",
                      0.0005, Unit::nanometerMinus2, 10 /* decimals */, 0.0001 /* step */,
                      RealLimits::nonnegative(), "density");

    m_interference.init("Interference function", "");
}

double ParticleLayoutItem::totalDensityValue() const
{
    if (!totalDensityIsDefinedByInterference())
        return m_ownDensity.value();

    ASSERT(m_interference.currentItem());

    if (const auto* interLatticeItem =
            dynamic_cast<const Interference2DAbstractLatticeItem*>(m_interference.currentItem())) {
        Lattice2DItem* latticeItem = interLatticeItem->latticeTypeItem();
        try {
            const double area = latticeItem->unitCellArea();
            return area == 0.0 ? 0.0 : 1.0 / area;
        } catch (const std::exception&) {
            // nothing to do here; new exception will be caught during job execution
            return 0.0;
        }
    }

    if (const auto* hd = dynamic_cast<const InterferenceHardDiskItem*>(m_interference.currentItem()))
        return hd->density();

    ASSERT(false);
}

QVector<ItemWithParticles*> ParticleLayoutItem::particles() const
{
    return m_particles;
}

void ParticleLayoutItem::addParticle(ItemWithParticles* particle)
{
    m_particles << particle;
}

void ParticleLayoutItem::removeParticle(ItemWithParticles* particle)
{
    m_particles.removeAll(particle);
    delete particle;
}

QVector<ItemWithParticles*> ParticleLayoutItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    for (auto* p : m_particles)
        result << p << p->containedItemsWithParticles();
    return result;
}

bool ParticleLayoutItem::totalDensityIsDefinedByInterference() const
{
    return dynamic_cast<const Interference2DAbstractLatticeItem*>(m_interference.currentItem())
           || dynamic_cast<const InterferenceHardDiskItem*>(m_interference.currentItem());
}

void ParticleLayoutItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::OwnDensity, m_ownDensity);
    m_interference.rwSelected(s, Tag::InterferenceFunction);
    Serialize::rwCatalogized<ItemWithParticlesCatalog>(s, Tag::Particles, m_particles, m_materials);
}
