//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ProfileItems.cpp
//! @brief     Implements Profile1DItem's classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ProfileItems.h"
#include "Base/Const/Units.h"
#include "GUI/Support/XML/Serialize.h"

namespace {
namespace Tag {

const QString Omega("Omega");
const QString OmegaX("OmegaX");
const QString OmegaY("OmegaY");
const QString Eta("Eta");
const QString Gamma("Gamma");

} // namespace Tag
} // namespace

Profile1DItem::Profile1DItem()
{
    m_omega.init("Omega", "Half-width of the distribution", 1.0, Unit::nanometer, "omega");
}

void Profile1DItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Omega, m_omega);
}

DoublePropertyRefs Profile1DItem::profileProperties()
{
    return {m_omega};
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile1D> Profile1DCauchyItem::createProfile() const
{
    return std::make_unique<Profile1DCauchy>(m_omega.value());
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile1D> Profile1DGaussItem::createProfile() const
{
    return std::make_unique<Profile1DGauss>(m_omega.value());
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile1D> Profile1DGateItem::createProfile() const
{
    return std::make_unique<Profile1DGate>(m_omega.value());
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile1D> Profile1DTriangleItem::createProfile() const
{
    return std::make_unique<Profile1DTriangle>(m_omega.value());
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile1D> Profile1DCosineItem::createProfile() const
{
    return std::make_unique<Profile1DCosine>(m_omega.value());
}

// --------------------------------------------------------------------------------------------- //

Profile1DVoigtItem::Profile1DVoigtItem()
{
    m_eta.init("Eta", "Parameter [0,1] to balance between Cauchy (eta=0.0) and Gauss (eta=1.0)",
               0.5, Unit::unitless, 3 /* decimals */, RealLimits::limited(0.0, 1.0), "eta");
}

std::unique_ptr<IProfile1D> Profile1DVoigtItem::createProfile() const
{
    return std::make_unique<Profile1DVoigt>(m_omega.value(), m_eta.value());
}

void Profile1DVoigtItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Omega, m_omega);
    Serialize::rwProperty(s, Tag::Eta, m_eta);
}

DoublePropertyRefs Profile1DVoigtItem::profileProperties()
{
    return Profile1DItem::profileProperties() + DoublePropertyRefs{m_eta};
}

// --------------------------------------------------------------------------------------------- //

Profile2DItem::Profile2DItem()
{
    m_omegaX.init("OmegaX", "Half-width of the distribution along its x-axis", 1.0, Unit::nanometer,
                  "omegaX");
    m_omegaY.init("OmegaY", "Half-width of the distribution along its y-axis", 1.0, Unit::nanometer,
                  "omegaY");
    m_gamma.init(
        "Gamma",
        "Angle in direct space between first lattice vector and x-axis of the distribution", 0.0,
        Unit::degree, "gamma");
}

void Profile2DItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::OmegaX, m_omegaX);
    Serialize::rwProperty(s, Tag::OmegaY, m_omegaY);
    Serialize::rwProperty(s, Tag::Gamma, m_gamma);
}

DoublePropertyRefs Profile2DItem::profileProperties()
{
    return {m_omegaX, m_omegaY, m_gamma};
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile2D> Profile2DCauchyItem::createProfile() const
{
    return std::make_unique<Profile2DCauchy>(m_omegaX.value(), m_omegaY.value(),
                                             Units::deg2rad(m_gamma.value()));
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile2D> Profile2DGaussItem::createProfile() const
{
    return std::make_unique<Profile2DGauss>(m_omegaX.value(), m_omegaY.value(),
                                            Units::deg2rad(m_gamma.value()));
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile2D> Profile2DGateItem::createProfile() const
{
    return std::make_unique<Profile2DGate>(m_omegaX.value(), m_omegaY.value(),
                                           Units::deg2rad(m_gamma.value()));
}

// --------------------------------------------------------------------------------------------- //

std::unique_ptr<IProfile2D> Profile2DConeItem::createProfile() const
{
    return std::make_unique<Profile2DCone>(m_omegaX.value(), m_omegaY.value(),
                                           Units::deg2rad(m_gamma.value()));
}

// --------------------------------------------------------------------------------------------- //

Profile2DVoigtItem::Profile2DVoigtItem()
{
    m_eta.init("Eta", "Parameter [0,1] to balance between Cauchy (eta=0.0) and Gauss (eta=1.0)",
               0.5, Unit::unitless, 3 /* decimals */, RealLimits::limited(0.0, 1.0), "eta");
}

std::unique_ptr<IProfile2D> Profile2DVoigtItem::createProfile() const
{
    return std::make_unique<Profile2DVoigt>(m_omegaX.value(), m_omegaY.value(),
                                            Units::deg2rad(m_gamma.value()), m_eta.value());
}

void Profile2DVoigtItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::OmegaX, m_omegaX);
    Serialize::rwProperty(s, Tag::OmegaY, m_omegaY);
    Serialize::rwProperty(s, Tag::Gamma, m_gamma);
    Serialize::rwProperty(s, Tag::Eta, m_eta);
}

DoublePropertyRefs Profile2DVoigtItem::profileProperties()
{
    return {m_omegaX, m_omegaY, m_eta, m_gamma};
}
