//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Common/DataItemBundleWidget.h
//! @brief     Defines class DataItemBundleWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_COMMON_DATAITEMBUNDLEWIDGET_H
#define BORNAGAIN_GUI_VIEW_COMMON_DATAITEMBUNDLEWIDGET_H

#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include <QWidget>

//! The DataItemBundleWidget class is a base for widgets, representing the derivatives of DataItem,
//! namely SpecularDataItem and IntensityDataItem. It stores a list of items that are required for
//! simultaneous, synchronous work while plotting and changing their properties.

class DataItemBundleWidget : public QWidget {
    Q_OBJECT
public:
    explicit DataItemBundleWidget(QWidget* parent = nullptr);
    ~DataItemBundleWidget() override;

    virtual void setDataItem(DataItem* dataItem);
    virtual void setDataItems(const QList<DataItem*>& dataItems);

    //! Access to non-null specular and intensity items
    QList<SpecularDataItem*> specularDataItems() const;
    QList<IntensityDataItem*> intensityDataItems() const;

    //! Access to first specular and intensity items
    SpecularDataItem* currentSpecularDataItem() const;
    IntensityDataItem* currentIntensityDataItem() const;

private:
    template <typename T>
    QList<T*> tItems() const;
    template <typename T>
    T* tFirstItem() const;

    QList<DataItem*> m_dataItems;
};

template <typename T>
QList<T*> DataItemBundleWidget::tItems() const
{
    QList<T*> output;
    for (DataItem* item : m_dataItems) {
        T* tItem = dynamic_cast<T*>(item);
        if (tItem)
            output.push_back(tItem);
    }
    return output;
}

template <typename T>
T* DataItemBundleWidget::tFirstItem() const
{
    T* item = dynamic_cast<T*>(m_dataItems.first());
    return item;
}

#endif // BORNAGAIN_GUI_VIEW_COMMON_DATAITEMBUNDLEWIDGET_H
