//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/Pyramid4.cpp
//! @brief     Implements class Pyramid4.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/Pyramid4.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"

const ff::Topology Pyramid4::topology = {{{{3, 2, 1, 0}, true},
                                          {{0, 1, 5, 4}, false},
                                          {{1, 2, 6, 5}, false},
                                          {{2, 3, 7, 6}, false},
                                          {{3, 0, 4, 7}, false},
                                          {{4, 5, 6, 7}, true}},
                                         false};

Pyramid4::Pyramid4(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_base_edge(m_P[0])
    , m_height(m_P[1])
    , m_alpha(m_P[2])
{
    validateOrThrow();
}

Pyramid4::Pyramid4(double base_edge, double height, double alpha)
    : Pyramid4(std::vector<double>{base_edge, height, alpha})
{
}

std::string Pyramid4::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_base_edge, "base_edge");
    requestGt0(errs, m_height, "height");
    const double cot_alpha = Math::cot(m_alpha);
    if (m_alpha <= 0 || !std::isfinite(cot_alpha) || cot_alpha < 0)
        errs.push_back("pyramid angle alpha " + std::to_string(m_alpha) + " out of bounds");
    if (!errs.empty())
        return jointError(errs);

    double r = cot_alpha * 2 * m_height / m_base_edge; // L(top)/L(base)
    if (r > 1)
        return jointError({"parameters violate condition 2*ctg(alpha) <= base_edge/height"});

    double a = m_base_edge / 2;
    double b = a * (1 - r);

    // center of mass
    double zcom = m_height * (1. / 4) * (6 - 8 * r + 3 * r * r) / (3 - 3 * r + r * r);

    setPolyhedron(topology, -zcom,
                  {// base:
                   {-a, -a, -zcom},
                   {a, -a, -zcom},
                   {a, a, -zcom},
                   {-a, a, -zcom},
                   // top:
                   {-b, -b, m_height - zcom},
                   {b, -b, m_height - zcom},
                   {b, b, m_height - zcom},
                   {-b, b, m_height - zcom}});

    m_validated = true;
    return "";
}
