#include "Base/Axis/FixedBinAxis.h"
#include "Base/Const/Units.h"
#include "Base/Math/Constants.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Tests/GTestWrapper/google_test.h"

class DepthProbeCoordinatesTest : public ::testing::Test {
protected:
    DepthProbeCoordinatesTest();

    void checkMainFunctionality(const DepthProbeCoordinates& test_object);
    void checkAlphaAxis(Coords units, const DepthProbeCoordinates& test_object);
    void checkZAxis(Coords units, const DepthProbeCoordinates& test_object);
    const double m_alpha_start = 0.5; // first axis value in rads
    const double m_alpha_end = 1.0;   // last axis value in rads
    const double m_z_start = -30.0;
    const double m_z_end = 10.0;
    const size_t m_nbins = 100;
    Beam m_beam;
    std::vector<const IAxis*> m_axes;
};

DepthProbeCoordinatesTest::DepthProbeCoordinatesTest()
    : m_beam(Beam::horizontalBeam())
    , m_axes({new FixedBinAxis("Angles", m_nbins, m_alpha_start, m_alpha_end), // angles in radians
              new FixedBinAxis("Positions", m_nbins, m_z_start, m_z_end)})     // z positions in nm
{
}

void DepthProbeCoordinatesTest::checkMainFunctionality(const DepthProbeCoordinates& test_object)
{
    EXPECT_EQ(test_object.rank(), 2u);

    EXPECT_NEAR(test_object.calculateMin(0, Coords::UNDEFINED), 2.8647889757e+1,
                2.8647889757e+1 * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::DEGREES), 2.8647889757e+1,
                2.8647889757e+1 * 1e-10);
    EXPECT_NEAR(test_object.calculateMin(0, Coords::QSPACE), 6.0246390001, 6.0246390001 * 1e-10);
    EXPECT_EQ(test_object.calculateMin(0, Coords::RADIANS), m_alpha_start);
    EXPECT_EQ(test_object.calculateMin(0, Coords::NBINS), 0.0);

    EXPECT_NEAR(test_object.calculateMax(0, Coords::UNDEFINED), 5.7295779513e+1,
                5.7295779513e+1 * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::DEGREES), 5.7295779513e+1,
                5.7295779513e+1 * 1e-10);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::QSPACE), 1.0574236256e+1,
                1.0574236256e+1 * 1e-10);
    EXPECT_EQ(test_object.calculateMax(0, Coords::RADIANS), m_alpha_end);
    const auto n_bins = static_cast<double>(m_nbins);
    EXPECT_NEAR(test_object.calculateMax(0, Coords::NBINS), n_bins, n_bins * 1e-10);

    checkAlphaAxis(Coords::UNDEFINED, test_object);
    checkAlphaAxis(Coords::DEGREES, test_object);
    checkAlphaAxis(Coords::RADIANS, test_object);
    checkAlphaAxis(Coords::QSPACE, test_object);
    checkAlphaAxis(Coords::NBINS, test_object);

    checkZAxis(Coords::UNDEFINED, test_object);
    checkZAxis(Coords::DEGREES, test_object);
    checkZAxis(Coords::RADIANS, test_object);
    checkZAxis(Coords::QSPACE, test_object);
    checkZAxis(Coords::NBINS, test_object);
}

void DepthProbeCoordinatesTest::checkAlphaAxis(Coords units,
                                               const DepthProbeCoordinates& test_object)
{
    std::unique_ptr<IAxis> axis(test_object.createConvertedAxis(0, units));
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis.get()));
    EXPECT_EQ(axis->size(), test_object.axisSize(0));
    EXPECT_EQ(axis->size(), m_nbins);
    EXPECT_EQ(axis->min(), test_object.calculateMin(0, units));
    EXPECT_EQ(axis->max(), test_object.calculateMax(0, units));
}

void DepthProbeCoordinatesTest::checkZAxis(Coords units, const DepthProbeCoordinates& test_object)
{
    std::unique_ptr<IAxis> axis(test_object.createConvertedAxis(1, units));
    EXPECT_TRUE(dynamic_cast<const FixedBinAxis*>(axis.get()));
    EXPECT_EQ(axis->size(), test_object.axisSize(1));
    EXPECT_EQ(axis->size(), m_nbins);

    EXPECT_EQ(axis->min(), test_object.calculateMin(1, units));
    const double test_min = units == Coords::NBINS ? 0 : m_z_start;
    EXPECT_NEAR(axis->min(), test_min, std::abs(test_min) * 1e-10);

    EXPECT_EQ(axis->max(), test_object.calculateMax(1, units));
    const double test_max = units == Coords::NBINS ? m_nbins : m_z_end;
    EXPECT_NEAR(axis->max(), test_max, std::abs(test_max) * 1e-10);
}

TEST_F(DepthProbeCoordinatesTest, DepthProbeCoordinates)
{
    DepthProbeCoordinates converter(m_axes, m_beam.direction(), m_beam.wavelength());
    checkMainFunctionality(converter);
}

TEST_F(DepthProbeCoordinatesTest, DepthProbeCoordinatesExceptions)
{
    DepthProbeCoordinates converter(m_axes, m_beam.direction(), m_beam.wavelength());

    EXPECT_FAILED_ASSERT(converter.axisName(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.axisName(1, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.axisName(2, Coords::UNDEFINED));

    EXPECT_FAILED_ASSERT(converter.calculateMin(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMin(1, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMin(2, Coords::UNDEFINED));

    EXPECT_FAILED_ASSERT(converter.calculateMax(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMax(1, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.calculateMax(2, Coords::RADIANS));

    EXPECT_FAILED_ASSERT(converter.createConvertedAxis(0, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.createConvertedAxis(1, Coords::MM));
    EXPECT_FAILED_ASSERT(converter.createConvertedAxis(2, Coords::UNDEFINED));
}

TEST_F(DepthProbeCoordinatesTest, DepthProbeCoordinatesClone)
{
    DepthProbeCoordinates converter(m_axes, m_beam.direction(), m_beam.wavelength());
    std::unique_ptr<DepthProbeCoordinates> converter_clone(converter.clone());
    checkMainFunctionality(*converter_clone);
}
